/*
 * Copyright (c) 2008-2010 Lu, Chao-Ming (Tetralet).  All rights reserved.
 * 
 * This file is part of LilyTerm.
 *
 * LilyTerm is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * LilyTerm is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with LilyTerm.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "dialog.h"
#define TEMPSTR 6

extern GtkWidget *menu_active_window;
extern struct ModKey modkeys[MOD];
extern struct Page_Color page_color[PAGE_COLOR];
extern GList* window_list;
extern gchar *profile_dir;
extern gchar *key_groups[KEY_GROUP];
extern struct KeyValue system_keys[KEYS];
extern gchar *init_LC_CTYPE;

gboolean dialog_actived = FALSE;
gboolean force_to_quit = FALSE;
gboolean using_kill = FALSE;

// CHANGE_TAB_NAME,
// CHANGE_THE_FOREGROUND_COLOR,
// ADJUST_THE_BRIGHTNESS_OF_ANSI_COLORS_USED_IN_TERMINAL,
// ADJUST_THE_BRIGHTNESS_OF_ANSI_COLORS_WHEN_INACTIVE,
// CHANGE_THE_BACKGROUND_COLOR,
// CHANGE_BACKGROUND_SATURATION,
// CHANGE_THE_OPACITY_OF_WINDOW,
// CHANGE_THE_TEXT_COLOR_OF_WINDOW_TITLE,
// CHANGE_THE_TEXT_COLOR_OF_CMDLINE,
// CHANGE_THE_TEXT_COLOR_OF_CURRENT_DIR,
// CHANGE_THE_TEXT_COLOR_OF_CUSTOM_PAGE_NAME,
// CHANGE_THE_TEXT_COLOR_OF_ROOT_PRIVILEGES_CMDLINE,
// CHANGE_THE_TEXT_COLOR_OF_NORMAL_TEXT,
// CONFIRM_TO_EXECUTE_COMMAND,
// CONFIRM_TO_CLOSE_MULTI_PAGES,
// CONFIRM_TO_CLOSE_RUNNING_APPLICATION,
// CONFIRM_TO_CLOSE_A_TAB_WITH_CHILD_PROCCESS,
// CONFIRM_TO_CLOSE_A_WINDOW_WITH_CHILD_PROCCESS,
// CONFIRM_TO_EXIT_WITH_CHILD_PROCCESS,
// SET_FUNCTION_KEY_VALUE,
// USAGE_MESSAGE


gboolean dialog(GtkWidget *widget, gsize style)
{
#ifdef DETAIL
	g_debug("! Launch dialog() with style = %ld", (long)style);
#endif
	dialog_actived = TRUE;

	struct Dialog *dialog = g_new0(struct Dialog, 1);
	struct Color_Data *color_data = g_new0(struct Color_Data, 1);
	color_data->original_page_color = NULL;

	struct Window *win_data = NULL;
	// A critical error message for current_vte = NULL.

#ifdef DEBUG
	// menu_active_window==NULL: it should NOT happen!
	if (menu_active_window==NULL)
	{
		gchar *err_msg = g_strdup_printf("dialog(%ld): menu_active_window = NULL\n\n"
					  "Please report bug to %s, Thanks!",
					  (glong)style, PACKAGE_BUGREPORT);
		error_dialog(NULL, _("The following error is occurred:"),
			     "The following error is occurred:",
			     GTK_STOCK_DIALOG_ERROR, err_msg, NULL);
		g_free(err_msg);
		return FALSE;
	}
	else
#endif
		win_data = (struct Window *)g_object_get_data(G_OBJECT(menu_active_window), "Win_Data");
	// g_debug("dialog(): win_data->runtime_LC_MESSAGES = %s", win_data->runtime_LC_MESSAGES);
	setlocale(LC_MESSAGES, win_data->runtime_LC_MESSAGES);

	struct Page *page_data = NULL;
	if (win_data!=NULL)
	{
		if (win_data->current_vte!=NULL)
		{
			page_data = (struct Page *)g_object_get_data(G_OBJECT(win_data->current_vte), "Page_Data");

			// g_debug("Get win_data = %p, page_data = %p, current_vte = %p when creating dialog!",
			//	win_data, page_data, win_data->current_vte);

			// For changing the text color of tab
			dialog->total_page = gtk_notebook_get_n_pages(GTK_NOTEBOOK(
										page_data->notebook));
			dialog->current_page_no = gtk_notebook_get_current_page(GTK_NOTEBOOK(
										page_data->notebook));
		}
	}

	gboolean tab_bar_is_shown = TRUE;
	gboolean response = TRUE;
	gchar *temp_str[TEMPSTR] = {NULL};
	gint create_entry_hbox = 0;
	gint i;

	// g_debug("Running dialog mod%d\n", style);

	/*
		/--------------main_hbox--------------\
		|/icon_vbox\/------state_vbox-----\/-\|
		||	   ||/----title_hbox-----\||r||
		||	   |||			 |||i||
		||	   ||\-------------------/||g||
		||	   ||/----entry_hbox-----\||h||
		||	   |||			 |||t||
		||	   |||			 ||| ||
		||	   ||\-------------------/||v||
		||	   ||/-state_bottom_hbox-\||b||
		||	   |||			  ||o||
		||	   ||\-------------------/||x||
		|\---------/\---------------------/\-/|
		|     /------------\ /------------\   |
		|     |	    OK	   | |	 CANCEL	  |   |
		|     \------------/ \------------/   |
		\-------------------------------------/
	*/

	switch (style)
	{
		case CHANGE_TAB_NAME:						// 1
			create_dialog(_("Rename this tab"),
				      "Rename this tab",
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      FALSE,
				      10,
				      GTK_RESPONSE_OK,
				      GTK_STOCK_DIALOG_INFO,
				      _("Please input the Label Name of this tab:"),
				      FALSE,
				      0,
				      TRUE,
				      1,
				      dialog);

			dialog->label = gtk_label_new (_("Label Name:"));
			gtk_box_pack_start (GTK_BOX(dialog->entry_hbox), dialog->label, FALSE, FALSE, 0);
			// entry area
			dialog->entry = gtk_entry_new ();
			if (page_data->custom_page_name == NULL)
				gtk_entry_set_text(GTK_ENTRY(dialog->entry), page_data->page_name);
			else
				gtk_entry_set_text(GTK_ENTRY(dialog->entry), page_data->custom_page_name);
			gtk_box_pack_start (GTK_BOX (dialog->entry_hbox), dialog->entry, TRUE, TRUE, 0);
			// set activate = TRUE when [enter] pressed in entry
			gtk_entry_set_activates_default(GTK_ENTRY(dialog->entry), TRUE);

			break;
		case CHANGE_BACKGROUND_SATURATION:				// 2
		{
			create_dialog(_("Change the saturation of background"),
				      "Change the saturation of background",
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      TRUE,
				      10,
				      GTK_RESPONSE_OK,
				      GTK_STOCK_DIALOG_INFO,
				      _("Change the Saturation of background:"),
				      FALSE,
				      0,
				      TRUE,
				      2,
				      dialog);

			create_scale_widget(dialog, 0, 1, 0.001,
					    win_data->background_saturation,
					    (GSourceFunc)set_background_saturation,
					    win_data->current_vte);

			dialog->original_transparent_background = win_data->transparent_background;
			if (! compare_strings (win_data->background_image, NULL_DEVICE))
				win_data->transparent_background = TRUE;
			set_background_saturation(NULL, 0, win_data->background_saturation,
						  win_data->current_vte);

			break;
		}
#ifdef ENABLE_GDKCOLOR_TO_STRING
		case CHANGE_THE_FOREGROUND_COLOR:				// 9
			create_dialog(_("Change the foreground color in terminal"),
				      "Change the foreground color in terminal",
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      FALSE,
				      15,
				      GTK_RESPONSE_OK,
				      NULL,
				      NULL,
				      FALSE,
				      0,
				      FALSE,
				      0,
				      dialog);
			color_data->original_color = win_data->fg_color;
			create_color_selection_widget(dialog, color_data, style,
						      (GSourceFunc)adjest_vte_color, win_data->current_vte);
			break;
		case CHANGE_THE_TEXT_COLOR_OF_WINDOW_TITLE:			// 11
		case CHANGE_THE_TEXT_COLOR_OF_CMDLINE:				// 12
		case CHANGE_THE_TEXT_COLOR_OF_CURRENT_DIR:			// 13
		case CHANGE_THE_TEXT_COLOR_OF_CUSTOM_PAGE_NAME:			// 14
		case CHANGE_THE_TEXT_COLOR_OF_ROOT_PRIVILEGES_CMDLINE:		// 15
		case CHANGE_THE_TEXT_COLOR_OF_NORMAL_TEXT:			// 16
		{
			init_locale_restrict_data(win_data->runtime_LC_MESSAGES);
			gint page_color_type = style - CHANGE_THE_TEXT_COLOR_OF_WINDOW_TITLE;
			temp_str[0] = g_strdup_printf(_("Change the %s color on tab"),
						      page_color[page_color_type].comment);
			temp_str[1] = g_strdup_printf("Change the %s color on tab",
						      page_color[page_color_type].comment);
			create_dialog(temp_str[0],
				      temp_str[1],
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      FALSE,
				      15,
				      GTK_RESPONSE_OK,
				      NULL,
				      NULL,
				      FALSE,
				      0,
				      FALSE,
				      0,
				      dialog);
			// save the dialog data first
			g_object_set_data(G_OBJECT(dialog->window), "Dialog", dialog);

			color_data->original_page_color = win_data->user_page_color[page_color_type];
			gdk_color_parse(win_data->user_page_color[page_color_type], &(color_data->original_color));

			// vte = the 1st page of current notebook.
			struct Page *page_data = get_page_data_from_nth_page(win_data, 0);
			create_color_selection_widget(dialog, color_data, style,
						      (GSourceFunc)adjest_vte_color, page_data->vte);
			// store the is_bold data of 1st page...
			dialog->tab_1_is_bold = page_data->is_bold;
			// win_data->lost_focus = FALSE;

			// only bold the 1st page
			gboolean is_bold = TRUE;
			// the color for using in 1st page, can NOT be free()!
			gchar *current_color = NULL;
			// the text for using in 1st page, can NOT be free()!
			gchar *display_text = NULL;

			// add new pages to PAGE_COLOR+1.
			// g_debug("dialog->total_page = %d, current_page = %d",
			//	dialog->total_page, gtk_notebook_get_current_page(GTK_NOTEBOOK(win_data->notebook)));
			// The new pages will added next to 1st page!
			for (i=dialog->total_page; i<=PAGE_COLOR; i++)
				add_page(win_data,
					 page_data,
					 NULL,
					 NULL,
					 NULL,
					 NULL,
					 NULL,
					 NULL,
					 FALSE,
					 NULL,
					 FALSE);
			// we need to set the current page to 1st page after adding page.
			gtk_notebook_set_current_page(GTK_NOTEBOOK(win_data->notebook), 0);

			for (i=0; i<=PAGE_COLOR; i++)
			{
				// ste the page name and color for demo.
				struct Page *page_data = get_page_data_from_nth_page(win_data, i);
				dialog->original_update_method[i] = page_data->page_update_method;
				// g_debug ("??? Save: (%d) page_data->page_update_method = %d", i+1, page_data->page_update_method);
				if (i)
				{
					display_text = _(page_color[i-1].comment);
					current_color = win_data->user_page_color[i-1];
				}
				else
				{
					display_text = _("Bold Demo Text");
					current_color = color_data->original_page_color;
				}

				// get the demo_vte and demo_text.
				// the change of color will be shown immediately in 1st page and the page here.
				if ((i+CHANGE_THE_TEXT_COLOR_OF_WINDOW_TITLE-1)==style)
				{
					color_data->demo_vte = page_data->vte;
					color_data->demo_text = display_text;
				}
				// g_debug("!!! updating %d page name!", i+1);
				update_page_name(page_data->window,
						 page_data->vte,
						 NULL,
						 page_data->label_text,
						 i+1,
						 display_text,
						 current_color,
						 FALSE,
						 is_bold,
						 FALSE,
						 NULL,
						 page_data->custom_window_title,
						 FALSE);
				// only 1st page is bold.
				is_bold = FALSE;
			}
			tab_bar_is_shown = gtk_notebook_get_show_tabs GTK_NOTEBOOK(win_data->notebook);
			if (! tab_bar_is_shown)
				hide_and_show_tabs_bar(win_data, TRUE);

			break;
		}
#else
		case CHANGE_THE_FOREGROUND_COLOR:
		case CHANGE_THE_TEXT_COLOR_OF_WINDOW_TITLE:			// 11
		case CHANGE_THE_TEXT_COLOR_OF_CMDLINE:				// 12
		case CHANGE_THE_TEXT_COLOR_OF_CURRENT_DIR:			// 13
		case CHANGE_THE_TEXT_COLOR_OF_CUSTOM_PAGE_NAME:			// 14
		case CHANGE_THE_TEXT_COLOR_OF_ROOT_PRIVILEGES_CMDLINE:		// 15
		case CHANGE_THE_TEXT_COLOR_OF_NORMAL_TEXT:			// 16
			return upgrade_dialog(ENABLE_GDKCOLOR_TO_STRING_VER);
			break;
#endif
		case CONFIRM_TO_CLOSE_MULTI_PAGES:				// 3
			if (g_list_length(window_list) > 1) create_entry_hbox = 2;
			create_dialog(_("Confirm to close multi tabs"),
				      "Confirm to close multi tabs",
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      FALSE,
				      10,
				      GTK_RESPONSE_OK,
				      GTK_STOCK_DIALOG_QUESTION,
				      _("You are about to close multi tabs. Continue anyway?"),
				      FALSE,
				      0,
				      TRUE,
				      create_entry_hbox,
				      dialog);
			create_SIGKILL_and_EXIT_widget(dialog, FALSE, create_entry_hbox, _("those tabs"));

			break;
		case SET_FUNCTION_KEY_VALUE:					// 4
		{
			init_locale_restrict_data(win_data->runtime_LC_MESSAGES);
			temp_str[0] = g_markup_escape_text(_("You may press <Ctrl + Alt + Shift> + <any key> here\n"
							     "to get the key value used for the function key."), -1);
			create_dialog(_("Get the key value for function key"),
				      "Get the key value for function key",
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      FALSE,
				      10,
				      GTK_RESPONSE_OK,
				      GTK_STOCK_DIALOG_INFO,
				      temp_str[0],
				      FALSE,
				      0,
				      TRUE,
				      2,
				      dialog);

			// The Key Group frame

			// Copy the key values
			for (i=0; i<KEYS; i++)
				dialog->user_key_value[i] = g_strdup(win_data->user_keys[i].value);
			// The Key Group frame
			GtkTreeStore *store = gtk_tree_store_new (1, G_TYPE_STRING);
			GtkTreeIter key_group_tree_iter[KEY_GROUP], tmp_tree_iter;
			gint key_index[KEYS]= {0};
			for (i=0; i<KEY_GROUP; i++)
			{
				gtk_tree_store_append (store, &(key_group_tree_iter[i]), NULL);
				gtk_tree_store_set (store, &(key_group_tree_iter[i]), 0, key_groups[i], -1);
			}
			for (i=0; i<KEYS; i++)
			{
				gtk_tree_store_append (store, &tmp_tree_iter, &(key_group_tree_iter[system_keys[i].group]));
				gtk_tree_store_set (store, &tmp_tree_iter, 0, system_keys[i].topic, -1);
				dialog->KeyTree[system_keys[i].group][key_index[system_keys[i].group]] = i;
				// g_debug("Set KeyTree[%d][%d] = %d", system_keys[i].group,
				//				    key_index[system_keys[i].group],
				//				    i);
				key_index[system_keys[i].group]++;
			}
			
			// Show the combo box
			GtkTreeModel *model= GTK_TREE_MODEL(store);
			GtkWidget *combo = gtk_combo_box_new_with_model(model);
			g_object_unref (model);
			GtkCellRenderer *renderer = gtk_cell_renderer_text_new (); 
			gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (combo), renderer, TRUE);
			gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (combo), renderer, "text", 0, NULL);
			// Hide the capital
			gtk_cell_layout_set_cell_data_func (GTK_CELL_LAYOUT (combo), renderer,
							    hide_combo_box_capital, NULL, NULL);
			// Set the default value
			GtkTreeIter iter;
			GtkTreePath *path = gtk_tree_path_new_from_indices (0, 0, -1);
			gtk_tree_model_get_iter (model, &iter, path);
			gtk_tree_path_free (path);
			gtk_combo_box_set_active_iter (GTK_COMBO_BOX (combo), &iter);
			g_signal_connect(G_OBJECT(combo), "changed", G_CALLBACK(update_key_info), dialog);

			// Add the combo box to frame
			GtkWidget *hbox = gtk_hbox_new (FALSE, 0);
			GtkWidget *label = gtk_label_new(_("Key Group: "));
			gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, 0);
			dialog->adjustment = gtk_label_new(NULL);
			set_markup_key_value(NULL, key_groups[KEY_GROUP_MISC], dialog->adjustment);
			gtk_box_pack_start (GTK_BOX(hbox), dialog->adjustment, FALSE, FALSE, 0);
			GtkWidget *vbox = create_frame_widget(dialog, NULL, hbox, combo, 0);

			hbox = gtk_hbox_new (FALSE, 0);
			label = gtk_label_new(_("Note: "));
			gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, 0);
			dialog->label = gtk_label_new(NULL);
			set_markup_key_value("dark green", system_keys[KEY_DISABLE_FUNCTION].translation, dialog->label);
			gtk_box_pack_start (GTK_BOX(hbox), dialog->label, FALSE, FALSE, 0);
			gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

			// The Key Value frame
			dialog->key_value_label = gtk_label_new(NULL);
			if ((dialog->user_key_value[KEY_DISABLE_FUNCTION]==NULL) ||
			    (dialog->user_key_value[KEY_DISABLE_FUNCTION][0]=='\0'))
				set_markup_key_value("blue",
						     _("(Disabled)"),
						     dialog->key_value_label);
			else
				set_markup_key_value("blue",
						     dialog->user_key_value[KEY_DISABLE_FUNCTION],
						     dialog->key_value_label);
			gtk_label_set_selectable(GTK_LABEL(dialog->key_value_label), TRUE);
			create_frame_widget(dialog, _("Key Value:"), NULL, dialog->key_value_label, 0);
			g_signal_connect_after( G_OBJECT(dialog->window), "key-press-event",
						G_CALLBACK(dialog_key_press), dialog);

			label = create_button_with_image(_("Disable all the function keys of this group."),
							 GTK_STOCK_CLOSE, (GSourceFunc)clear_key_group, dialog);
			gtk_box_pack_start (GTK_BOX(dialog->entry_hbox), label, TRUE, TRUE, 0);
			label = create_button_with_image(_("Disable all the function keys."),
							 GTK_STOCK_CLOSE, (GSourceFunc)clear_key_group_all, dialog);
			gtk_box_pack_start (GTK_BOX(dialog->entry_hbox), label, TRUE, TRUE, 0);
			break;
		}
		case USAGE_MESSAGE:						// 5
		{
			temp_str[0] = g_strdup_printf(_("Usage for %s %s"),PACKAGE_NAME, VERSION);
			temp_str[1] = g_strdup_printf("Usage for %s %s",PACKAGE_NAME, VERSION);
			// gchar *help_msg = get_help_message(win_data->profile);
			temp_str[2] = get_help_message_usage(win_data->profile, TRUE);

			create_dialog(temp_str[0],
				      temp_str[1],
				      DIALOG_OK,
				      page_data->window,
				      FALSE,
				      FALSE,
				      15,
				      GTK_RESPONSE_OK,
				      NULL,
				      temp_str[2],
				      TRUE,
				      0,
				      FALSE,
				      2,
				      dialog);
			/* Create the expander */
			temp_str[2] = convert_text_to_html(_("Default shortcut key: (It may custom or disable by right click menu [Set function key value])"), FALSE, NULL, "tt", NULL);
			GtkWidget *expander = gtk_expander_new (temp_str[2]);
			gtk_expander_set_use_markup (GTK_EXPANDER(expander), TRUE);
			gtk_box_pack_start (GTK_BOX (dialog->entry_hbox), expander, FALSE, FALSE, 0);
			temp_str[3] = get_help_message_function_key(TRUE);
			GtkWidget *label = gtk_label_new (NULL);
			gtk_label_set_markup (GTK_LABEL(label), temp_str[3]);
			gtk_expander_set_spacing (GTK_EXPANDER(expander), 15);
			gtk_container_add (GTK_CONTAINER (expander), label);
			// g_object_set(label, "can-focus", FALSE, NULL);
#ifdef USE_GTK_WIDGET_SET_CAN_FOCUS
			gtk_widget_set_can_focus(GTK_WIDGET (expander), FALSE);
#else
			GTK_WIDGET_UNSET_FLAGS(expander, GTK_CAN_FOCUS);
#endif
			GtkWidget *hbox = gtk_hbox_new (FALSE, 0);
			gtk_box_pack_start (GTK_BOX(dialog->entry_hbox), hbox, FALSE, FALSE, 0);
			temp_str[4] = g_strdup_printf(_("\nPlease report bug to %s. Thank you for using %s!"),
						      PACKAGE_BUGREPORT, PACKAGE_NAME);
			temp_str[5] = convert_text_to_html (temp_str[4], FALSE, NULL, "tt", NULL);
			label = gtk_label_new (NULL);
			gtk_label_set_markup (GTK_LABEL (label), temp_str[5]);
			gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, 0);
			break;
		}
		case CONFIRM_TO_CLOSE_RUNNING_APPLICATION:			// 7
		{
			// get the command line of running foreground program
			pid_t tpgid = get_tpgid(page_data->pid);
			temp_str[0] = get_cmdline(tpgid);
			temp_str[1] = g_strdup_printf("(%d) %s", tpgid, temp_str[0]);
			temp_str[2] = g_markup_escape_text(temp_str[1], -1);
			temp_str[3] = g_strdup_printf(_("There is still a running foreground program on #%d tab:"),
							 page_data->page_no+1);
			temp_str[4] = g_strdup_printf(_("Terminate it by using `kill -9 %d' is recommend "
							    "if it is non-response."), page_data->pid);
			temp_str[5] = g_strconcat(temp_str[3],
						  "\n\n\t<b><span foreground=\"blue\">",
						  temp_str[2],
						  "</span></b>\n\n",
						  _("Continue anyway?"),
						  "\n\n<span foreground=\"darkred\">",
						  _("Tip:"),
						  " ",
						  temp_str[4],
						  "</span>",
						  NULL);
		
			if ((g_list_length(window_list) > 1) || (dialog->total_page>1)) create_entry_hbox = 2;

			create_dialog(_("Confirm to close running foreground program"),
				      "Confirm to close running foreground program",
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      FALSE,
				      10,
				      GTK_RESPONSE_OK,
				      GTK_STOCK_DIALOG_QUESTION,
				      temp_str[5],
				      TRUE,
				      70,
				      TRUE,
				      create_entry_hbox,
				      dialog);

			create_SIGKILL_and_EXIT_widget(dialog, FALSE, create_entry_hbox, _("this tab"));
			break;
		}
		case CHANGE_THE_OPACITY_OF_WINDOW:				// 8
#ifdef ENABLE_RGBA
			create_dialog(_("Change the opacity of window"),
				      "Change the opacity of window",
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      TRUE,
				      10,
				      GTK_RESPONSE_OK,
				      GTK_STOCK_DIALOG_INFO,
				      _("Change the Opacity of window:"),
				      FALSE,
				      0,
				      TRUE,
				      2,
				      dialog);
			create_scale_widget(dialog, 0, 1, 0.001,
					    win_data->window_opacity, 
					    (GSourceFunc)set_window_opacity,
					    page_data->window);
			dialog->original_transparent_window = win_data->transparent_window;
			dialog->original_window_opacity = win_data->window_opacity;
			win_data->transparent_window = TRUE;
			set_window_opacity(NULL, 0, win_data->window_opacity, page_data->window);
#else
		return upgrade_dialog(ENABLE_RGBA_VER);
#endif
			break;
		case CHANGE_THE_BACKGROUND_COLOR:				// 10
			create_dialog(_("Change the background color in terminal"),
				      "Change the background color in terminal",
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      FALSE,
				      15,
				      GTK_RESPONSE_OK,
				      NULL,
				      NULL,
				      FALSE,
				      0,
				      FALSE,
				      0,
				      dialog);
			color_data->original_color = win_data->bg_color;
			create_color_selection_widget(dialog, color_data, style,
						      (GSourceFunc)adjest_vte_color, win_data->current_vte);
			break;
		case ADJUST_THE_BRIGHTNESS_OF_ANSI_COLORS_USED_IN_TERMINAL:	// 19
		case ADJUST_THE_BRIGHTNESS_OF_ANSI_COLORS_WHEN_INACTIVE:	// 26
		{
			g_object_set_data(G_OBJECT(win_data->current_vte), "Dialog", dialog);
			dialog->original_using_custom_color = win_data->using_custom_color;
			win_data->using_custom_color = TRUE;
			dialog->original_color_brightness = win_data->color_brightness;
			dialog->original_fg_color = win_data->fg_color;

			gchar *dialog_title = NULL, *dialog_title_trans = NULL, *title = NULL;
			gdouble value;
			if (style == ADJUST_THE_BRIGHTNESS_OF_ANSI_COLORS_USED_IN_TERMINAL)
			{
				dialog_title_trans = _("Adjust the brightness of ANSI colors used in terminal");
				dialog_title = _("Adjust the brightness of ANSI colors used in terminal");
				title = _("Adjust the brightness of ANSI colors used in terminal");
				value = win_data->color_brightness;
			}
			else
			{
				dialog_title_trans = _("Adjust the brightness of ANSI colors when inactive");
				dialog_title = _("Adjust the brightness of ANSI colors when inactive");
				title = _("Adjust the brightness of ANSI colors when inactive");
				value = win_data->color_brightness_inactive;
				win_data->fg_color = get_inactive_color (win_data->fg_color,
									 win_data->color_brightness_inactive,
									 win_data->color_brightness);
				win_data->color_brightness = win_data->color_brightness_inactive;

				set_new_ansi_color(win_data);
			}
			create_dialog(dialog_title_trans,
				      dialog_title,
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      TRUE,
				      10,
				      GTK_RESPONSE_OK,
				      GTK_STOCK_DIALOG_INFO,
				      title,
				      FALSE,
				      0,
				      TRUE,
				      2,
				      dialog);
			create_scale_widget(dialog, -1, 1, 0.001,
					    value,
					    (GSourceFunc)set_ansi_color,
					    win_data->current_vte);
			break;
		}
		case CONFIRM_TO_EXECUTE_COMMAND:				// 22
			// g_debug("error_data.string = %s", error_data.string);
			// g_debug("error_data.encoding = %s", error_data.encoding);
			temp_str[0] = g_markup_escape_text(win_data->temp_data, -1);
			temp_str[1] = g_strconcat(PACKAGE_NAME,
						  _(" is trying to execute the following command:"),
						  "\n\n\t<b><span foreground=\"",
						  "blue",
						  "\">",
						  temp_str[0],
						  "</span></b>\n\n",
						  _("Continue anyway?"),
						  NULL);

			create_dialog(_("Confirm to execute -e/-x/--execute command"),
				      "Confirm to execute -e/-x/--execute command",
				      DIALOG_OK_CANCEL,
				      menu_active_window,
				      TRUE,
				      FALSE,
				      10,
				      GTK_RESPONSE_CANCEL,
				      GTK_STOCK_DIALOG_WARNING,
				      temp_str[1],
				      FALSE,
				      70,
				      TRUE,
				      0,
				      dialog);
			break;
		case CONFIRM_TO_CLOSE_A_TAB_WITH_CHILD_PROCCESS:		// 23
			// get the command line of running foreground program
			temp_str[0] = g_strdup_printf(_("The following programs are still running under #%d tab:"),
							 page_data->page_no+1);
			temp_str[1] = g_markup_escape_text(_("Using <Ctrl><D> or 'exit' to close tabs is recommended"),
							     -1);
			temp_str[2] = g_markup_escape_text(win_data->temp_data, -1);
			temp_str[3] = g_strconcat(temp_str[0],
						  "\n\n<b><span foreground=\"blue\">",
						  temp_str[2],
						  "</span></b>\n",
						  _("Continue anyway?"),
						  "\n\n<span foreground=\"darkred\">",
						  _("Tip:"),
						  " ",
						  temp_str[1],
						  "</span>",
						  NULL);

			create_dialog(_("Confirm to close a tab with child proccess"),
				      "Confirm to close a tab with child proccess",
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      FALSE,
				      10,
				      GTK_RESPONSE_CANCEL,
				      GTK_STOCK_DIALOG_QUESTION,
				      temp_str[3],
				      TRUE,
				      70,
				      TRUE,
				      2,
				      dialog);
		
			create_entry_hbox = ((g_list_length(window_list) > 1) || (dialog->total_page>1));
			create_SIGKILL_and_EXIT_widget(dialog, TRUE, create_entry_hbox, _("this tab"));

			break;
		case CONFIRM_TO_CLOSE_A_WINDOW_WITH_CHILD_PROCCESS:		// 24
			temp_str[0] = g_markup_escape_text(win_data->temp_data, -1);
			temp_str[1] = g_strconcat(_("The following programs are still running under this window:"),
						  "\n\n<b><span foreground=\"blue\">",
						  temp_str[0],
						  "</span></b>\n",
						  _("Continue anyway?"),
						  NULL);
			create_dialog(_("Confirm to close a window with child proccess"),
				      "Confirm to close a window with child proccess",
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      FALSE,
				      10,
				      GTK_RESPONSE_CANCEL,
				      GTK_STOCK_DIALOG_QUESTION,
				      temp_str[1],
				      TRUE,
				      70,
				      TRUE,
				      2,
				      dialog);
			create_entry_hbox = (g_list_length(window_list) > 1);
			create_SIGKILL_and_EXIT_widget(dialog, TRUE, create_entry_hbox, _("those tabs"));
			break;
		case CONFIRM_TO_EXIT_WITH_CHILD_PROCCESS:			// 25
			temp_str[0]  = g_strdup_printf(_("Confirm to close %s"), PACKAGE_NAME);
			temp_str[1]  = g_strdup_printf("Confirm to close %s", PACKAGE_NAME);
			// get the command line of running foreground program
			temp_str[2] = g_strdup_printf(_("The following programs are still running under %s:"), 
						      PACKAGE_NAME);
			temp_str[3] = g_markup_escape_text(win_data->temp_data, -1);
			temp_str[4] = g_strconcat(temp_str[2] ,
						  "\n\n<b><span foreground=\"blue\">",
						  temp_str[3],
						  "</span></b>\n",
						  _("Continue anyway?"),
						  NULL);
			create_dialog(temp_str[0],
				      temp_str[1],
				      DIALOG_OK_CANCEL,
				      page_data->window,
				      FALSE,
				      FALSE,
				      10,
				      GTK_RESPONSE_CANCEL,
				      GTK_STOCK_DIALOG_QUESTION,
				      temp_str[4],
				      TRUE,
				      70,
				      TRUE,
				      2,
				      dialog);
			create_SIGKILL_and_EXIT_widget(dialog, TRUE, TRUE, PACKAGE_NAME);

			break;
	}

	for (i=0; i<TEMPSTR; i++) g_free(temp_str[i]);

	// We'll use gtk_dialog_run() to show the whole dialog
	// g_debug("Finish creating dialog!");
	gtk_widget_show_all (dialog->main_hbox);

	if (gtk_dialog_run(GTK_DIALOG(dialog->window)) == GTK_RESPONSE_OK)
	{
		switch (style)
		{
			// style 1: change the tab's name
			case CHANGE_TAB_NAME:
				// g_debug("Got label name: %s\n",gtk_entry_get_text(GTK_ENTRY(entry)));
				g_free(page_data->custom_page_name);
				if (strlen(gtk_entry_get_text(GTK_ENTRY(dialog->entry))))
					// store page_data->page_name
					page_data->custom_page_name = g_strdup(gtk_entry_get_text(GTK_ENTRY(dialog->entry)));
				else
					page_data->custom_page_name = NULL;
			
				// g_debug("Get the page name = %s, color = %s",
				// 	page_data->custom_page_name, page_data->tab_color);
				if (page_data->page_update_method == 1)
					page_data->window_title_updated = 1;
				get_and_update_page_name(page_data, FALSE);

				break;
			// style  2: change the saturation of background
			case CHANGE_BACKGROUND_SATURATION:
			// style  9: change the foreground color
			case CHANGE_THE_FOREGROUND_COLOR:
			// style 10: change the background color
			case CHANGE_THE_BACKGROUND_COLOR:
			// style 11: change the text color of window title
			case CHANGE_THE_TEXT_COLOR_OF_WINDOW_TITLE:
			// style 12: change the text color of cmdline
			case CHANGE_THE_TEXT_COLOR_OF_CMDLINE:
			// style 13: change the text color of current dir
			case CHANGE_THE_TEXT_COLOR_OF_CURRENT_DIR:
			// style 14: change the text color of custom page name
			case CHANGE_THE_TEXT_COLOR_OF_CUSTOM_PAGE_NAME:
			// style 15: change the text color of root privileges cmdline
			case CHANGE_THE_TEXT_COLOR_OF_ROOT_PRIVILEGES_CMDLINE:
			// style 16: change the text color of normal text
			case CHANGE_THE_TEXT_COLOR_OF_NORMAL_TEXT:
			{
				// g_debug("Setting the colors. Type = %d", style);
				switch (style)
				{
					case CHANGE_BACKGROUND_SATURATION:
						win_data->background_saturation = gtk_range_get_value(GTK_RANGE(dialog->adjustment)) + 0.0005;
						break;
#ifdef ENABLE_GDKCOLOR_TO_STRING
					case CHANGE_THE_FOREGROUND_COLOR:
						g_free(win_data->foreground_color);
						win_data->foreground_color = 
							gdk_color_to_string(&(color_data->fg_color));
						win_data->fg_color = color_data->fg_color;
						adjust_ansi_color_severally(&(win_data->fg_color_inactive),
									    &(win_data->fg_color),
									    win_data->color_brightness_inactive / 
									    	(1 + win_data->color_brightness));
						break;
					case CHANGE_THE_BACKGROUND_COLOR:
						g_free(win_data->background_color);
						win_data->background_color =
							gdk_color_to_string(&(color_data->bg_color));
						win_data->bg_color = color_data->bg_color;
						// FIXME: GtkColorSelection have no ALPHA-CHANGED signal.
						//	  so that the following code should be
						//	  marked for temporary.
						// background_saturation = 
						//	gtk_color_selection_get_current_alpha(
						//		GTK_COLOR_SELECTION(adjustment))/65535;
						break;
					case CHANGE_THE_TEXT_COLOR_OF_WINDOW_TITLE:
					case CHANGE_THE_TEXT_COLOR_OF_CMDLINE:
					case CHANGE_THE_TEXT_COLOR_OF_CURRENT_DIR:
					case CHANGE_THE_TEXT_COLOR_OF_CUSTOM_PAGE_NAME:
					case CHANGE_THE_TEXT_COLOR_OF_ROOT_PRIVILEGES_CMDLINE:
					case CHANGE_THE_TEXT_COLOR_OF_NORMAL_TEXT:
					{
						gint page_color_type = style - CHANGE_THE_TEXT_COLOR_OF_WINDOW_TITLE;
						g_free(win_data->user_page_color[page_color_type]);
						win_data->user_page_color[page_color_type] = gdk_color_to_string(&(color_data->original_color));
						break;
					}
#endif
				}
				if (style>=CHANGE_THE_TEXT_COLOR_OF_WINDOW_TITLE &&
				    style<=CHANGE_THE_TEXT_COLOR_OF_NORMAL_TEXT)
					// recover the color of pages
					recover_page_colors(dialog->window, page_data->window, page_data->notebook);
				else
				{
					struct Page *tmp_page_data = NULL;
					// change the saturation of background and foreground/background color
					for (i=0; i<gtk_notebook_get_n_pages(GTK_NOTEBOOK(win_data->notebook)); i++)
					{
						tmp_page_data = get_page_data_from_nth_page(win_data, i);
						if (style==CHANGE_BACKGROUND_SATURATION)
							set_background_saturation (NULL, 0, 
										   win_data->background_saturation,
										   tmp_page_data->vte);

						if (style==CHANGE_THE_FOREGROUND_COLOR ||
						    style==CHANGE_THE_BACKGROUND_COLOR)
							adjest_vte_color (GTK_COLOR_SELECTION(dialog->adjustment),
									  tmp_page_data->vte);
					}
				}
				break;
			}
			// style 3: confirm to close multi pages
			case CONFIRM_TO_CLOSE_MULTI_PAGES:
			{
				// we need to destroy the dialog before confirm a working vte.
				if ((create_entry_hbox) && gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(dialog->label)))
				{
					gtk_main_quit();
					force_to_quit = TRUE;
					// g_debug("Set force_to_quit to %d", force_to_quit);
				}

				gtk_widget_destroy(dialog->window);
				dialog->window = NULL;
				close_multi_tabs(win_data, FALSE);
			
				break;
			}
			// style 4: get function key value
			case SET_FUNCTION_KEY_VALUE:
			{
				for (i=0; i<KEYS; i++)
					convert_string_to_user_key(i, dialog->user_key_value[i], win_data);
				break;
			}
			// style 23: confirm to close a tab with child proccess
			case CONFIRM_TO_CLOSE_A_TAB_WITH_CHILD_PROCCESS:
			// style 24: confirm to close a window with child proccess
			case CONFIRM_TO_CLOSE_A_WINDOW_WITH_CHILD_PROCCESS:
			// style 25: confirm to close LilyTerm with child proccess
			case CONFIRM_TO_EXIT_WITH_CHILD_PROCCESS:
				if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(dialog->key_value_label)))
					using_kill = TRUE;
				if ((style==CONFIRM_TO_CLOSE_A_WINDOW_WITH_CHILD_PROCCESS) || 
				    (style==CONFIRM_TO_EXIT_WITH_CHILD_PROCCESS))
					force_to_quit = TRUE;
			// style  7: confirm to close running application
			case CONFIRM_TO_CLOSE_RUNNING_APPLICATION:
				if ((create_entry_hbox) && (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(dialog->label))))
				{
					gtk_main_quit();
					force_to_quit = TRUE;
				}
				break;
#ifdef ENABLE_RGBA
			// style 8: change the opacity of window
			case CHANGE_THE_OPACITY_OF_WINDOW:
				win_data->window_opacity =
					gtk_range_get_value(GTK_RANGE(dialog->adjustment))+0.0005;
				break;
			// style 19: Adjust the brightness of ANSI colors used in terminal
			case ADJUST_THE_BRIGHTNESS_OF_ANSI_COLORS_USED_IN_TERMINAL:
			{
				struct Page *page_data;
				for (i=0; i<gtk_notebook_get_n_pages(GTK_NOTEBOOK(win_data->notebook)); i++)
				{
					page_data = get_page_data_from_nth_page(win_data, i);
					set_vte_color(win_data, page_data);
				}
				break;
			}
			case ADJUST_THE_BRIGHTNESS_OF_ANSI_COLORS_WHEN_INACTIVE:
				win_data->color_brightness_inactive = win_data->color_brightness;
				win_data->color_brightness = dialog->original_color_brightness;
				win_data->fg_color = dialog->original_fg_color;
				set_new_ansi_color(win_data);
				adjust_ansi_color(win_data->color_inactive,
						  win_data->color_orig,
						  win_data->color_brightness_inactive);
				break;
#endif
		}
	}
	else
	{
		switch (style)
		{
			// style  2: change the saturation of background
			case CHANGE_BACKGROUND_SATURATION:
			// style  9: change the foreground color
			case CHANGE_THE_FOREGROUND_COLOR:
			// style 10: change the background color
			case CHANGE_THE_BACKGROUND_COLOR:
				if (style==CHANGE_BACKGROUND_SATURATION)
				{
					win_data->transparent_background = dialog->original_transparent_background;
					// g_debug("transparent_background = %d", transparent_background);
					set_background_saturation(NULL,
								  0,
								  win_data->background_saturation,
								  win_data->current_vte);
				}
				if (style==CHANGE_THE_FOREGROUND_COLOR)
					color_data->fg_color = color_data->original_color;
				else if (style==CHANGE_THE_BACKGROUND_COLOR)
					color_data->bg_color = color_data->original_color;

				if (style!=CHANGE_BACKGROUND_SATURATION)
				{
					color_data->recover = TRUE;
					adjest_vte_color(GTK_COLOR_SELECTION(dialog->adjustment),
									     win_data->current_vte);
				}
				break;
			// style 4: get function key value
			case SET_FUNCTION_KEY_VALUE:
			{
				for (i=0; i<KEYS; i++)
					g_free(dialog->user_key_value[i]);
				break;
			}

#ifdef ENABLE_RGBA
			// style  8: change the opacity of window
			case CHANGE_THE_OPACITY_OF_WINDOW:
				win_data->transparent_window = dialog->original_transparent_window;
				win_data->window_opacity = dialog->original_window_opacity;
				set_window_opacity( NULL, 0, win_data->window_opacity, page_data->window);
				break;
#endif
			// style 11: change the text color of window title
			case CHANGE_THE_TEXT_COLOR_OF_WINDOW_TITLE:
			// style 12: change the text color of cmdline
			case CHANGE_THE_TEXT_COLOR_OF_CMDLINE:
			// style 13: change the text color of current dir
			case CHANGE_THE_TEXT_COLOR_OF_CURRENT_DIR:
			// style 14: change the text color of custom page name
			case CHANGE_THE_TEXT_COLOR_OF_CUSTOM_PAGE_NAME:
			// style 15: change the text color of root privileges cmdline
			case CHANGE_THE_TEXT_COLOR_OF_ROOT_PRIVILEGES_CMDLINE:
			// style 16: change the text color of normal text
			case CHANGE_THE_TEXT_COLOR_OF_NORMAL_TEXT:
				recover_page_colors(dialog->window, page_data->window, page_data->notebook);
				break;
			case ADJUST_THE_BRIGHTNESS_OF_ANSI_COLORS_USED_IN_TERMINAL:
				// g_debug("win_data->using_custom_color = %d", win_data->using_custom_color);
				// g_debug("win_data->using_custom_color_orig = %d", win_data->using_custom_color_orig);
				// g_debug("dialog->original_color_brightness = %0.3f", dialog->original_color_brightness);
				win_data->color_brightness = dialog->original_color_brightness;
				win_data->fg_color = dialog->original_fg_color;
				adjust_ansi_color(win_data->color,
						  win_data->color_orig,
						  win_data->color_brightness);
				struct Page *page_data = (struct Page *)g_object_get_data(G_OBJECT(win_data->current_vte),
											  "Page_Data");
				win_data->using_custom_color = dialog->original_using_custom_color;
				set_vte_color(win_data, page_data);
				break;
			case ADJUST_THE_BRIGHTNESS_OF_ANSI_COLORS_WHEN_INACTIVE:
				win_data->using_custom_color = dialog->original_using_custom_color;
				win_data->color_brightness = dialog->original_color_brightness;
				win_data->fg_color = dialog->original_fg_color;
				set_new_ansi_color(win_data);
				break;

		}
		response = FALSE;
	}
	// destroy dialog.
	if (dialog->window) gtk_widget_destroy(dialog->window);

	g_free(dialog);
	g_free(color_data);

	if ((style >= CHANGE_THE_TEXT_COLOR_OF_WINDOW_TITLE) && (style <= CHANGE_THE_TEXT_COLOR_OF_NORMAL_TEXT))
		hide_and_show_tabs_bar(win_data, tab_bar_is_shown);

	dialog_actived = FALSE;
	return response;
}

void create_dialog(gchar *dialog_title_translation, gchar *dialog_title,  Dialog_Button_Type type,
		   GtkWidget *window, gboolean center, gboolean resizable, gint border_width,
		   gint response, gchar *icon, gchar *title, gboolean selectable, gint max_width_chars,
		   gboolean state_bottom, gint create_entry_hbox, struct Dialog *dialog)
{
#ifdef DETAIL
	g_debug("! Launch create_dialog() with dialog_title_translation = %s, " 
		"dialog_title = %s, type = %d, window = %p, center = %d, resizable = %d, "
		"border_width = %d, response = %d, icon = %s, title = %s, selectable = %d, "
		"max_width_chars = %d, state_bottom = %d, create_entry_hbox = %d, dialog = %p",
		dialog_title_translation, dialog_title, type, window, center, resizable,  
		border_width, response, icon, title, selectable, max_width_chars, state_bottom,  
		create_entry_hbox, dialog);
#endif
	gboolean BOTTON_ORDER = gtk_alternative_dialog_button_order(NULL);
	// g_debug("gtk_alternative_dialog_button_order = %d" ,gtk_alternative_dialog_button_order (NULL));

	// Strange behavior, If the locale is empty, or setted to 'C' or 'POSIX',
	// And 'default_locale' is setted in the profile,
	// All the UI will be translated to default_locale,
	// But the window title will be empty.
	// So, set it to un-translated string here.
	if ((init_LC_CTYPE==NULL) || (init_LC_CTYPE[0]=='\0') ||
	    compare_strings(init_LC_CTYPE, "C") || compare_strings(init_LC_CTYPE, "\"C\"") ||
	    compare_strings(init_LC_CTYPE, "POSIX") || compare_strings(init_LC_CTYPE, "\"POSIX\""))
		dialog_title_translation = dialog_title;

	switch (type)
	{
		case DIALOG_OK:
			dialog->window = gtk_dialog_new_with_buttons (dialog_title_translation,
								      GTK_WINDOW(window),
								      GTK_DIALOG_NO_SEPARATOR |
									GTK_DIALOG_DESTROY_WITH_PARENT,
								      GTK_STOCK_OK,
								      GTK_RESPONSE_OK,
								      NULL);
			break;
		case DIALOG_OK_CANCEL:
			if (BOTTON_ORDER)
				dialog->window = gtk_dialog_new_with_buttons (dialog_title_translation,
									      GTK_WINDOW(window),
									      GTK_DIALOG_NO_SEPARATOR |
										GTK_DIALOG_DESTROY_WITH_PARENT,
									      GTK_STOCK_OK,
									      GTK_RESPONSE_OK,
									      GTK_STOCK_CANCEL,
									      GTK_RESPONSE_CANCEL,
									      NULL);
			else
				dialog->window = gtk_dialog_new_with_buttons (dialog_title_translation,
									      GTK_WINDOW(window),
									      GTK_DIALOG_NO_SEPARATOR |
										GTK_DIALOG_DESTROY_WITH_PARENT,
									      GTK_STOCK_CANCEL,
									      GTK_RESPONSE_CANCEL,
									      GTK_STOCK_OK,
									      GTK_RESPONSE_OK,
									      NULL);
			break;
	}

	// FIXME: It may be a bug in gtk+2?
	// if a dialog shown before main window is shown,
	// destory the dialog will destory the data of the icon, too.
	// and when showing main window with gtk_widget_show_all(win_data->window) later,
	// It will have no icon, and showing following error:
	// Gtk-CRITICAL **: gtk_window_realize_icon: assertion `info->icon_pixmap == NULL' failed
	// So, we need to set the icon for the dialog here.
	set_window_icon(dialog->window);

	if (center) gtk_window_set_position (GTK_WINDOW (dialog->window), GTK_WIN_POS_CENTER);
	gtk_window_set_resizable (GTK_WINDOW (dialog->window), resizable);
	gtk_container_set_border_width (GTK_CONTAINER (dialog->window), border_width);
	gtk_dialog_set_default_response(GTK_DIALOG(dialog->window), response);

	dialog->main_hbox = gtk_hbox_new (FALSE, 5);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog->window)->vbox), dialog->main_hbox);
	dialog->main_right_vbox = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_end (GTK_BOX(dialog->main_hbox), dialog->main_right_vbox, FALSE, FALSE, 0);


	if (icon)
	{
		dialog->icon_vbox = gtk_vbox_new (FALSE, 30);
		gtk_box_pack_start (GTK_BOX(dialog->main_hbox), dialog->icon_vbox, FALSE, FALSE, 10);
		dialog->icon = gtk_image_new_from_stock (icon, GTK_ICON_SIZE_DIALOG);
		gtk_box_pack_start (GTK_BOX(dialog->icon_vbox), dialog->icon, FALSE, FALSE, 10);
	}

	dialog->state_vbox = gtk_vbox_new (FALSE, 15); 
	gtk_box_pack_start (GTK_BOX(dialog->main_hbox), dialog->state_vbox, TRUE, TRUE, 0);

	dialog->title_hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX(dialog->state_vbox), dialog->title_hbox, FALSE, FALSE, 0);

	if (title)
	{
		dialog->title = gtk_label_new (NULL);
		gtk_label_set_markup (GTK_LABEL (dialog->title), title);
		if (selectable) gtk_label_set_selectable(GTK_LABEL(dialog->title), TRUE);
		if (max_width_chars)
		{
			gtk_label_set_max_width_chars (GTK_LABEL(dialog->title), max_width_chars);
			gtk_label_set_ellipsize(GTK_LABEL(dialog->title), PANGO_ELLIPSIZE_MIDDLE);
		}
		gtk_box_pack_start (GTK_BOX(dialog->title_hbox), dialog->title, FALSE, FALSE, 0);
	}

	if (state_bottom)
	{
		dialog->state_bottom_hbox = gtk_hbox_new (FALSE, 3);
		gtk_box_pack_end (GTK_BOX(dialog->state_vbox), dialog->state_bottom_hbox, FALSE, FALSE, 0);
	}

	if (create_entry_hbox)
	{
		switch (create_entry_hbox)
		{
			case 1:
				dialog->entry_hbox = gtk_hbox_new (FALSE, 5);
				break;
			case 2:
				dialog->entry_hbox = gtk_vbox_new (FALSE, 5);
				break;
		}
		gtk_box_pack_start (GTK_BOX(dialog->state_vbox), dialog->entry_hbox, TRUE, TRUE, 10);
	}
}

GtkWidget *create_frame_widget (struct Dialog *dialog, gchar *label,
				GtkWidget *label_widget, GtkWidget *child, guint padding)
{
#ifdef DETAIL
	g_debug("! Launch create_frame_widget() with dialog = %p, label = %s, label_widget = %p, "
		"child = %p, padding = %d!", dialog, label, label_widget, child, padding);
#endif
	GtkWidget *frame = gtk_frame_new (label);
	if (label_widget) gtk_frame_set_label_widget (GTK_FRAME(frame), label_widget);
	gtk_box_pack_start (GTK_BOX(dialog->entry_hbox), frame, FALSE, FALSE, padding);
	GtkWidget *vbox = gtk_vbox_new (FALSE, 15);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 10);
	gtk_container_add (GTK_CONTAINER (frame), vbox);
	if (child) gtk_box_pack_start (GTK_BOX(vbox), child, FALSE, FALSE, 0);
	return vbox;
}

GtkWidget *create_button_with_image(gchar *label_text, const gchar *stock_id,
				    GtkFunction func, gpointer func_data)
{
#ifdef DETAIL
	g_debug("! Launch create_button_with_image() with label_text = %s, stock_id = %s!",
		label_text, stock_id);
#endif
	GtkWidget *label = gtk_button_new_with_label(label_text);
	gtk_button_set_image (GTK_BUTTON(label),
			      gtk_image_new_from_stock(stock_id, GTK_ICON_SIZE_MENU));
	gtk_button_set_relief(GTK_BUTTON(label), GTK_RELIEF_NONE);
	gtk_button_set_focus_on_click(GTK_BUTTON(label), FALSE); 
	gtk_button_set_alignment(GTK_BUTTON(label), 0, 0.5);
	g_signal_connect(G_OBJECT(label), "clicked", G_CALLBACK(func), func_data);
	return label;
}

void create_color_selection_widget(struct Dialog *dialog, struct Color_Data *color_data, Dialog_Type_Flags style,
				   GtkFunction func, gpointer func_data)
{
#ifdef DETAIL
	g_debug("! Launch create_color_selection_widget() with dialog = %p, color_data = %p, style = %d!",
		dialog, color_data, style);
#endif
	dialog->adjustment = gtk_color_selection_new();
	// save the color data first
	g_object_set_data(G_OBJECT(dialog->adjustment), "Color_Data", color_data);
	gtk_color_selection_set_has_opacity_control(GTK_COLOR_SELECTION(dialog->adjustment),
						    FALSE);
	gtk_color_selection_set_has_palette(GTK_COLOR_SELECTION(dialog->adjustment), FALSE);
	color_data->type = style;
	color_data->recover = FALSE;

	// set the previous/current color of gtk_color_selection dialog
	gtk_color_selection_set_previous_color (GTK_COLOR_SELECTION(dialog->adjustment),
						&(color_data->original_color));
	gtk_color_selection_set_current_color ( GTK_COLOR_SELECTION(dialog->adjustment),
						&(color_data->original_color));

	gtk_box_pack_start (GTK_BOX(dialog->title_hbox), dialog->adjustment, TRUE, TRUE, 0);

	g_signal_connect_after(dialog->adjustment, "color-changed",
			       G_CALLBACK(func), func_data);
}

void create_scale_widget(struct Dialog *dialog, gdouble min, gdouble max, gdouble step, gdouble value,
			 GtkFunction func, gpointer func_data)
{
#ifdef DETAIL
	g_debug("! Launch create_scale_widget() with dialog = %p, min = %3f, max = %3f, step = %3f, value = %3f!",
		dialog, min, max, step, value);
#endif
	GtkWidget *hbox1 = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX(dialog->entry_hbox), hbox1, FALSE, FALSE, 10);
	dialog->adjustment = gtk_hscale_new_with_range(min, max, step);
	gtk_widget_set_size_request(dialog->adjustment, 210, -1);
	gtk_range_set_value(GTK_RANGE(dialog->adjustment), value);
	g_signal_connect_after(dialog->adjustment, "change-value", G_CALLBACK(func), func_data);
	gtk_box_pack_start (GTK_BOX(dialog->entry_hbox), dialog->adjustment, TRUE, TRUE, 0);
	GtkWidget *hbox2 = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_end (GTK_BOX(dialog->entry_hbox), hbox2, FALSE, FALSE, 5);
}

void create_SIGKILL_and_EXIT_widget(struct Dialog *dialog, gboolean create_entry_hbox,
				    gboolean create_force_kill_hbox, gchar *count_str)
{
#ifdef DETAIL
	g_debug("! Launch create_SIGKILL_and_EXIT_widget() with dialog = %p, create_entry_hbox = %d, "
		"create_force_kill_hbox = %d, count_str = %s!",
		dialog, create_entry_hbox, create_force_kill_hbox, count_str);
#endif
	if (create_entry_hbox)
	{
		gchar *message = g_strdup_printf(_("Try to kill %s with SIGKILL() "
					  "to avoid terminate those programs!\n"
					  "(MAY CAUSE DATA LOSS!!)"), count_str);
		dialog->key_value_label = gtk_check_button_new_with_label(message);
		g_free(message);
		gtk_box_pack_start (GTK_BOX(dialog->entry_hbox), dialog->key_value_label, FALSE, FALSE, 0);
#ifdef USE_GTK_WIDGET_SET_CAN_FOCUS
		gtk_widget_set_can_focus(GTK_WIDGET (dialog->key_value_label), FALSE);
#else
		GTK_WIDGET_UNSET_FLAGS(dialog->key_value_label, GTK_CAN_FOCUS);
#endif
	}

	if (create_force_kill_hbox)
	{
		gchar *exit_str = g_strdup_printf(_("Force to close all the tabs and windows, and exit %s!"),
						  PACKAGE_NAME);
		dialog->label = gtk_check_button_new_with_label(exit_str);
		g_free(exit_str);
		gtk_box_pack_start (GTK_BOX(dialog->entry_hbox), dialog->label, FALSE, FALSE, 0);
#ifdef USE_GTK_WIDGET_SET_CAN_FOCUS
		gtk_widget_set_can_focus(GTK_WIDGET (dialog->label), FALSE);
#else
		GTK_WIDGET_UNSET_FLAGS(dialog->label, GTK_CAN_FOCUS);
#endif
	}
}

// The returned string should be freed when no longer needed.
gchar *get_colorful_profile(struct Window *win_data) 
{
#ifdef DETAIL
	g_debug("! Launch get_colorful_profile() with win_data = %p", win_data);
#endif
	gchar *profile;
	if (win_data->use_custom_profile)
		profile = g_strdup_printf("%s", win_data->profile);
	else
		profile = g_strdup_printf("%s/%s", profile_dir, USER_PROFILE);

	// g_debug("win_data->runtime_encoding = %s", win_data->runtime_encoding);
	// g_debug("profile = %s", profile);
	gchar *utf8_profile = convert_str_to_utf8(profile, win_data->runtime_encoding);
	g_free(profile);
	// g_debug("utf8_profile = %s", utf8_profile);
	return convert_text_to_html(utf8_profile, TRUE, "blue", "b", NULL);
}

gboolean dialog_key_press(GtkWidget *widget, GdkEventKey *event, struct Dialog *dialog)
{
#ifdef DETAIL
	g_debug("! Launch dialog_key_press() with key value = %s", gdk_keyval_name(event->keyval));
#endif

	gchar *key_value = g_strdup("");
	gint i=0;

	// FIXME: Be careful for the setted key value (GDK_Shift_L abd GDK_Hyper_R) here...
	if (((event->state & SHIFT_ONLY_MASK) || (event->keyval > GDK_asciitilde)) &&
	    ((event->keyval < GDK_Shift_L) || (event->keyval > GDK_Hyper_R)))
	{
		for (i=0; i<MOD; i++)
			if (event->state & modkeys[i].mod)
				key_value = dialog_key_press_join_string(key_value, "+", modkeys[i].name);
	}

	g_free(dialog->user_key_value[dialog->current_key_index]);
	if (key_value[0]!='\0')
	{
		key_value = dialog_key_press_join_string(key_value, " ", gdk_keyval_name(event->keyval));
		dialog->user_key_value[dialog->current_key_index] = key_value;
	}
	else
	{
		dialog->user_key_value[dialog->current_key_index] = key_value;
		key_value = _("(Disabled)");
	}

	set_markup_key_value("blue", key_value, dialog->key_value_label);

	return TRUE;
}

void set_markup_key_value(gchar *color, gchar *key_value, GtkWidget *label)
{
#ifdef DETAIL
	g_debug("! Launch set_markup_key_value() whith color = %s, key_value = %s, label = %p!",
		color, key_value, label);
#endif
	gchar *markup_text = convert_text_to_html (key_value, FALSE, color, "b", NULL);
	gtk_label_set_markup (GTK_LABEL (label), markup_text);
	g_free(markup_text);
}

// The "key_value" will free() in dialog_key_press_join_string()
gchar *dialog_key_press_join_string(gchar *key_value, gchar *separator, gchar *append)
{
#ifdef DETAIL
	g_debug("! Launch dialog_key_press_join_string() with key_value = %s, separator = %s, append = %s", key_value, separator, append);
#endif

	gchar *join_string = NULL;
	// We will not check NULL value!
	if (key_value[0]=='\0')
		join_string = g_strdup(append);
	else
		join_string = g_strdup_printf("%s%s%s",key_value, separator, append);

	g_free(key_value);
	return join_string;
}

void adjest_vte_color(GtkColorSelection *colorselection, GtkWidget *vte)
{
#ifdef DETAIL
	g_debug("! Launch adjest_vte_color() with colorselection = %p, vte = %p", colorselection, vte);
#endif
	// g_debug("Changing the color for vte %p", vte);
	struct Color_Data *color_data = (struct Color_Data*)g_object_get_data(G_OBJECT(colorselection), "Color_Data");

	switch (color_data->type)
	{
		case CHANGE_THE_FOREGROUND_COLOR:
			if (! color_data->recover)
				gtk_color_selection_get_current_color (colorselection, &(color_data->fg_color));
			vte_terminal_set_color_foreground(VTE_TERMINAL(vte), &(color_data->fg_color));
			vte_terminal_set_color_bold (VTE_TERMINAL(vte), &(color_data->fg_color));
			break;
		case CHANGE_THE_BACKGROUND_COLOR:
			if (! color_data->recover)
			{
				gtk_color_selection_get_current_color (colorselection, &(color_data->bg_color));
				// FIXME: GtkColorSelection have no ALPHA CHANGED signal.
				//	  so that the following code should be marked for temporary
				//if (use_rgba)
				//{
				//	gint alpha = gtk_color_selection_get_current_alpha(colorselection);
				//	g_debug("Current alpha = %d", alpha);
				//	vte_terminal_set_opacity(VTE_TERMINAL(vte), alpha);
				//	vte_terminal_set_background_saturation( VTE_TERMINAL(vte), 1-alpha/65535);
				//}
			}
			//else if (use_rgba)
			//	set_background_saturation(NULL, 0, background_saturation, vte);
		
			vte_terminal_set_color_background(VTE_TERMINAL(vte), &(color_data->bg_color));
			vte_terminal_set_background_tint_color (VTE_TERMINAL(vte), &(color_data->bg_color));
			break;
		case CHANGE_THE_TEXT_COLOR_OF_WINDOW_TITLE:
		case CHANGE_THE_TEXT_COLOR_OF_CMDLINE:
		case CHANGE_THE_TEXT_COLOR_OF_CURRENT_DIR:
		case CHANGE_THE_TEXT_COLOR_OF_CUSTOM_PAGE_NAME:
		case CHANGE_THE_TEXT_COLOR_OF_ROOT_PRIVILEGES_CMDLINE:
		case CHANGE_THE_TEXT_COLOR_OF_NORMAL_TEXT:
		{
			// the change of color will be shown(demo) in 1st page and demo page.
			gchar *current_color = NULL;
			gtk_color_selection_get_current_color (colorselection, &(color_data->original_color));
#ifdef ENABLE_GDKCOLOR_TO_STRING
			current_color = gdk_color_to_string(&(color_data->original_color));
#endif
			struct Page *page_data = (struct Page *)g_object_get_data(G_OBJECT(vte), "Page_Data");
			update_page_name(page_data->window, vte, NULL, page_data->label_text, 1,
					  _("Bold Demo Text"), current_color, FALSE, TRUE, FALSE, NULL,
					  page_data->custom_window_title, FALSE);
		
			page_data = (struct Page *)g_object_get_data(G_OBJECT(color_data->demo_vte), "Page_Data");
			update_page_name(page_data->window, color_data->demo_vte, NULL, page_data->label_text,
					 page_data->page_no+1, color_data->demo_text, current_color, FALSE, FALSE,
					 FALSE, NULL, page_data->custom_window_title, FALSE);
			g_free(current_color);
			break;
		}
	}
}

void recover_page_colors(GtkWidget *dialog_window, GtkWidget *window, GtkWidget *notebook)
{
#ifdef DETAIL
	g_debug("! Launch recover_page_colors() in window %p", window);
#endif
	struct Dialog *dialog = (struct Dialog *)g_object_get_data(G_OBJECT(dialog_window), "Dialog");
	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(window), "Win_Data");

	gint i;
	struct Page *page_data = NULL;

	// g_debug("page_cmdline_color = %s", win_data->page_cmdline_color);
	// g_debug("page_dir_color = %s", win_data->page_dir_color);
	// g_debug("page_custom_color = %s", win_data->page_custom_color);
	// g_debug("page_root_color = %s", win_data->page_root_color);
	// g_debug("page_normal_color = %s", win_data->page_normal_color);

	// delete the tab that we add when demo
	win_data->kill_color_demo_vte = TRUE;
	for (i=PAGE_COLOR; i>=dialog->total_page; i--)
	{
		page_data = get_page_data_from_nth_page(win_data, i);
		close_page (page_data->vte, TRUE);
	}
	win_data->kill_color_demo_vte = FALSE;

	// reset to the current page
	gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), dialog->current_page_no);

	// recover the title/color of pages
	for (i=0; i<dialog->total_page; i++)
	{
		page_data = get_page_data_from_nth_page(win_data, i);

		// restore the is_bold of 1st page
		// g_debug("i = %d", i);
		if (i==0)
			page_data->is_bold = dialog->tab_1_is_bold;
	
		// g_debug("Launch get_and_update_page_name() in recover_page_colors() for %d page!", i);
		// g_debug("page_data->page_update_method = %d", page_data->page_update_method);
		if (i <= PAGE_COLOR) page_data->page_update_method = dialog->original_update_method[i];
		if (page_data->page_update_method == 1) page_data->window_title_updated = 1;
		get_and_update_page_name(page_data, FALSE);
	}
}

void create_invalid_locale_error_message(gchar *locale)
{
#ifdef DETAIL
	g_debug("! Launch create_invalid_locale_error_message() with locale = %s!", locale);
#endif
	gchar *err_msg = g_strdup_printf(_("The locale \"%s\" seems NOT supported by your system!"), locale);
	error_dialog(NULL, _("Not supported locale!"), "Not supported locale!",
		     GTK_STOCK_DIALOG_WARNING, err_msg, NULL);
	g_free(err_msg);
}

// 1. menu_active_window = NULL
// 2. Not supported feature
// 3. The format of socket data is out of date
// 4. The following settings can NOT be applied
// 5. Not supported locale!
// 6. Error when creating child process
// 7. Error when writing profile
// 8. The format of your profile is out of date
// 9. The profile is invalid

void error_dialog(GtkWidget *window, gchar *title_translation, gchar *title,
		  gchar *icon, gchar *message, gchar *encoding)
{
#ifdef DETAIL
	g_debug("! Launch error_dialog() with window = %p, title = %s, icon = %s, message = %s, encodig = %s",
		window, title, icon, message, encoding);
#endif
// void create_dialog(gchar *dialog_title, gint type, GtkWidget *window, gboolean center,
//		      gboolean resizable, gint border_width, gint response, gchar *icon,
//		      gchar *title, gboolean selectable, gint max_width_chars,
//		      gboolean state_bottom, gint create_entry_hbox, struct Dialog *dialog)

	dialog_actived = TRUE;

	gchar *utf8_message = convert_str_to_utf8(message, encoding);
	struct Dialog dialog;

	// void create_dialog(gchar *dialog_title, Dialog_Button_Type type, GtkWidget *window, gboolean center,
	//		      gboolean resizable, gint border_width, gint response, gchar *icon,
	//		      gchar *title, gboolean selectable, gint max_width_chars,
	//		      gboolean state_bottom, gint create_entry_hbox, struct Dialog *dialog)

	create_dialog(title_translation, title, DIALOG_OK, window, TRUE,
		      FALSE, 10, GTK_RESPONSE_OK, icon, utf8_message,
		      FALSE, 0, TRUE, 0, &dialog);
	
	g_free(utf8_message);
	gtk_widget_show_all (dialog.main_hbox);
	gtk_dialog_run(GTK_DIALOG(dialog.window));
	gtk_widget_destroy(dialog.window);
	dialog_actived = FALSE;
}

gboolean upgrade_dialog(gchar *version_str)
{
#ifdef DETAIL
	g_debug("! Launch upgrade_dialog() with version_str = %s", version_str);
#endif
	gchar *err_msg = g_strdup_printf(_("You should upgrade to %s to support this feature."), version_str);
	error_dialog(NULL, _("Not supported feature!"), "Not supported feature!",
		     GTK_STOCK_DIALOG_WARNING, err_msg, NULL);
	g_free(err_msg);
	return FALSE;
}

gboolean set_ansi_color(GtkRange *range, GtkScrollType scroll, gdouble value, GtkWidget *vte)
{
#ifdef DETAIL
	g_debug("! Launch set_ansi_color() with range = %p, scroll = %u, value = %f, vte = %p",
		range, scroll, value, vte);
#endif
	struct Page *page_data = (struct Page *)g_object_get_data(G_OBJECT(vte), "Page_Data");
	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(page_data->window),
								     "Win_Data");
	struct Dialog *dialog = (struct Dialog *)g_object_get_data(G_OBJECT(vte), "Dialog");

	// g_debug("Get win_data = %d when set background saturation!", win_data);
	value = CLAMP(value, -1, 1);
	win_data->color_brightness = value;
	win_data->fg_color = get_inactive_color(dialog->original_fg_color,
						win_data->color_brightness,
						dialog->original_color_brightness);
	adjust_ansi_color(win_data->color, win_data->color_orig, win_data->color_brightness);
	set_vte_color(win_data, page_data);
	return FALSE;
}


GdkColor get_inactive_color(GdkColor original_fg_color, gdouble new_brightness, gdouble old_brightness)
{
#ifdef DETAIL
	g_debug("! Launch get_inactive_color() with new_brightness = %3f, "
		"old_brightness = %3f!", new_brightness, old_brightness);
#endif
	GdkColor inactive_color;
	if (new_brightness < old_brightness)
		adjust_ansi_color_severally(&inactive_color,
					    &(original_fg_color),
					    (new_brightness - old_brightness) /
						(1 + old_brightness));
	else
		adjust_ansi_color_severally(&inactive_color,
					    &(original_fg_color),
					    (new_brightness - old_brightness) /
						(1 - old_brightness));
	return inactive_color;
}

void set_new_ansi_color(struct Window *win_data)
{
#ifdef DETAIL
	g_debug("! Launch set_new_ansi_color() with win_data = %p!", win_data);
#endif
	win_data->fg_color_inactive = get_inactive_color (win_data->fg_color,
							  win_data->color_brightness_inactive,
							  win_data->color_brightness);
	adjust_ansi_color(win_data->color,
			  win_data->color_orig,
			  win_data->color_brightness);
	struct Page *page_data = (struct Page *)g_object_get_data(G_OBJECT(win_data->current_vte),
								  "Page_Data");
	set_vte_color(win_data, page_data);
}

void hide_combo_box_capital(GtkCellLayout *cell_layout, GtkCellRenderer *cell,
			    GtkTreeModel *tree_model, GtkTreeIter *iter, gpointer data)
{
#ifdef DETAIL
	g_debug("! Launch hide_combo_box_capital() with cell_layout = %p, cell = %p, tree_model = %p, iter = %p!",
		cell_layout, cell, tree_model, iter);
#endif
	g_object_set (cell, "sensitive", !gtk_tree_model_iter_has_child (tree_model, iter), NULL);
}

void update_key_info (GtkComboBox *combo, struct Dialog *dialog)
{
#ifdef DETAIL
	g_debug("! Launch update_key_info() with combo = %p, dialog = %p!", combo, dialog);
#endif
	GtkTreeIter iter;
	if (gtk_combo_box_get_active_iter (combo, &iter))
	{
		GtkTreePath *path = gtk_tree_model_get_path(gtk_combo_box_get_model(combo), &iter);
		gchar *string = gtk_tree_path_to_string (path);
		gchar **indices = g_strsplit(string, ":", -1);
		if ((indices[0]==NULL) || (indices[1]==NULL)) return;
		dialog->current_key_index = dialog->KeyTree[atoi(indices[0])][atoi(indices[1])];
		// g_debug("KeyTree[%d][%d] = %d", atoi(indices[0]), atoi(indices[1]), dialog->current_key_index);
		set_markup_key_value("dark green", system_keys[dialog->current_key_index].translation, dialog->label);
		// g_debug("dialog->current_key_index = %d, "
		//	   "dialog->user_key_value[dialog->current_key_index] = %s",
		//	   dialog->current_key_index,
		//	   dialog->user_key_value[dialog->current_key_index]);
		gchar *user_key_value = dialog->user_key_value[dialog->current_key_index];
		if ((user_key_value==NULL) || (user_key_value[0]=='\0'))
			user_key_value = _("(Disabled)");
		set_markup_key_value("blue", user_key_value, dialog->key_value_label);
		gtk_tree_path_free(path);
		g_free(string);
		g_strfreev(indices);

		set_markup_key_value(NULL,
				     key_groups[system_keys[dialog->current_key_index].group],
				     dialog->adjustment);
	}
	return;
}

void clear_key_groups(struct Dialog *dialog, gboolean clear_all)
{
#ifdef DETAIL
	g_debug("! Launch clear_key_groups() with dialog = %p, clear_all = %d!", dialog, clear_all);
#endif
	gint i;
	for (i=0; i<KEYS; i++)
	{
		if (clear_all || (system_keys[i].group == system_keys[dialog->current_key_index].group))
		{
			g_free(dialog->user_key_value[i]);
			dialog->user_key_value[i] = g_strdup("");
		}
	}
	set_markup_key_value("blue", _("(Disabled)"), dialog->key_value_label);
}

void clear_key_group(GtkButton *button, struct Dialog *dialog)
{
#ifdef DETAIL
	g_debug("! Launch clear_key_group() with button = %p, dialog = %p!", button, dialog);
#endif
	clear_key_groups(dialog, FALSE);
}

void clear_key_group_all(GtkButton *button, struct Dialog *dialog)
{
#ifdef DETAIL
	g_debug("! Launch clear_key_group_all() with button = %p, dialog = %p!", button, dialog);
#endif
	clear_key_groups(dialog, TRUE);
}

//void err_page_data_is_null(gchar *function_name)
//{
//	gchar *err_msg = g_strdup_printf("%s: page_data = NULL\n\n"
//					 "Please report bug to %s, Thanks!",
//					 function_name,
//					 PACKAGE_BUGREPORT);
//	error_dialog(NULL, err_msg, NULL, ERROR_MESSAGES);
//#ifdef DETAIL
//	g_debug("* free err_msg %p (%s) in set_encoding()", err_msg, err_msg);
//#endif
//	g_free(err_msg);
//}
