/*
 * Copyright (c) 2008-2010 Lu, Chao-Ming (Tetralet).  All rights reserved.
 * 
 * This file is part of LilyTerm.
 *
 * LilyTerm is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * LilyTerm is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with LilyTerm.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "misc.h"
extern gboolean proc_exist;

// The returned string should be freed when no longer needed.
gchar *convert_array_to_string(gchar **array, gchar separator)
{
#ifdef DETAIL
	g_debug("! Launch convert_array_to_string()");
#endif
	GString *array_str = g_string_new (NULL);
	gint i=0;

	if (array!=NULL)
	{
		while (array[i]!=NULL)
		{
			// g_debug("%d: %s", i, array[i]);
			if (array_str->len)
				g_string_append_printf(array_str, "%c%s", separator, array[i]);
			else
				g_string_append_printf(array_str, "%s", array[i]);
			i++;
		}
	}
	// g_debug ("Got array_str = %s", array_str->str);
	return g_string_free(array_str, FALSE);
}

// return FALSE if the strings are the same.
gboolean compare_strings(const gchar *string_a, const gchar *string_b)
{
#ifdef DETAIL
	g_debug("! Launch compare_strings()! with string_a = %s, string_b = %s", string_a, string_b);
#endif

	if ((string_a == NULL) && (string_b == NULL)) return FALSE;
	if ((string_a == NULL) || (string_b == NULL)) return TRUE;
	if (strcmp(string_a, string_b)) return TRUE;
	return FALSE;
}

// The returned string should be freed when no longer needed.
gchar *convert_str_to_utf8(gchar *string, gchar *encoding_str)
{
#ifdef DETAIL
	g_debug("! Launch convert_str_to_utf8() with string = %s, encoding_str = %s", string, encoding_str);
#endif
	gchar *local_string = NULL;

	if (encoding_str)
	{
		//g_debug("string = %s", string);
		//g_debug("encoding_str = %s", encoding_str);
		local_string = g_convert_with_fallback (string,
							-1,
							"UTF-8",
							encoding_str,
							"_",
							NULL,
							NULL,
							NULL);
		// g_debug("Get local_string = %s", local_string);

		if ((local_string == NULL ) || (local_string[0]=='\0'))
		{
			g_free(local_string);
			local_string = g_strdup(string);
			gint i = 0;
			while (local_string[i])
			{
				if (local_string[i] < 32 || local_string[i]>126)
					local_string[i] = '?';
				i++;
			}
		}

		// g_debug("FINAL: local_string = %s", local_string);
		return local_string;
	}
	else
		return g_strdup(string);
}

void set_VTE_CJK_WIDTH_environ(gint VTE_CJK_WIDTH)
{
#ifdef DETAIL
	g_debug("! Launch set_VTE_CJK_WIDTH_environ() with VTE_CJK_WIDTH = %d", VTE_CJK_WIDTH);
#endif
	// Set the VTE_CJK_WIDTH environment
	switch (VTE_CJK_WIDTH)
	{
		case 0:
			// Don't touch VTE_CJK_WIDTH
			break;
		//case 1:
		//	// The VTE_CJK_WIDTH will be setted depend on the LC environ in newer libvte version.
		//	unsetenv("VTE_CJK_WIDTH");
		//	break;
		case 1:
			// VTE_CJK_WIDTH=narrow only work for vte >= 0.16.14
			setenv("VTE_CJK_WIDTH", "narrow", TRUE);
			break;
		case 2:
			setenv("VTE_CJK_WIDTH", "wide", TRUE);
			break;
	}
}

gchar *get_VTE_CJK_WIDTH_str(gint VTE_CJK_WIDTH)
{
#ifdef DETAIL
	g_debug("! Launch get_VTE_CJK_WIDTH_str() with VTE_CJK_WIDTH = %d", VTE_CJK_WIDTH);
#endif
	switch (VTE_CJK_WIDTH)
	{
		case 0:
			return NULL;
		case 1:
			return "narrow";
		case 2:
			return "wide";
	}
	return NULL;
}

gint get_default_VTE_CJK_WIDTH()
{
#ifdef DETAIL
	g_debug("! Launch get_default_VTE_CJK_WIDTH()");
#endif

	const gchar *VTE_CJK_WIDTH = g_getenv("VTE_CJK_WIDTH");
	if (VTE_CJK_WIDTH==NULL)
		return 0;
	else
	{
		// VTE_CJK_WIDTH only work under UTF-8
		if ((g_ascii_strcasecmp (VTE_CJK_WIDTH, "wide")==0) ||
		    (g_ascii_strcasecmp (VTE_CJK_WIDTH, "1")==0))
			return 2;
		else if ((g_ascii_strcasecmp (VTE_CJK_WIDTH, "narrow")==0) ||
			 (g_ascii_strcasecmp (VTE_CJK_WIDTH, "0")==0))
			return 1;
		else
			return 0;
	}
}

void restore_SYSTEM_VTE_CJK_WIDTH_STR()
{
#ifdef DETAIL
	g_debug("! Launch restore_SYSTEM_VTE_CJK_WIDTH_STR()");
#endif
	extern gchar *SYSTEM_VTE_CJK_WIDTH_STR;
	if (SYSTEM_VTE_CJK_WIDTH_STR)
		g_setenv("VTE_CJK_WIDTH", SYSTEM_VTE_CJK_WIDTH_STR, TRUE);
	else
		g_unsetenv("VTE_CJK_WIDTH");
}

void set_env(const gchar *variable, const gchar *value, gboolean overwrite)
{
#ifdef DETAIL
	g_debug("! Launch set_env() with variable = %s, value = %s, overwrite = %d", variable, value, overwrite);
#endif

	if (value)
		g_setenv(variable, value, overwrite);
	else
		g_unsetenv(variable);
}

// get default locale from environ
// The returned string CAN NOT be free()!
gchar *get_default_LC_DATA(gint LC_TYPE)
{
#ifdef DETAIL
	g_debug("! Launch get_default_LC_DATA() with LC_TYPE= %d", LC_TYPE);
#endif
	char *LC_DATA;
	switch (LC_TYPE)
	{
		case LC_CTYPE:
			LC_DATA = (char*)g_getenv("LC_CTYPE");
			break;
		case LC_MESSAGES:
			LC_DATA = (char*)g_getenv("LC_MESSAGES");
			break;
		default:
			return "";
	}
	char *LC_ALL = (char*)g_getenv("LC_ALL");
	char *LANG = (char*)g_getenv("LANG");

	if ( (!LC_DATA) && LANG)
		LC_DATA = LANG;

	if (LC_ALL)
		LC_DATA = LC_ALL;

//	g_debug("Get LC_DATA = %s", LC_DATA);
	if (LC_DATA)
		return LC_DATA;
	else
		return "";
}

// The returned string should be freed when no longer needed.
gchar *get_encoding_from_locale(const gchar *locale)
{
#ifdef DETAIL
	g_debug("! Launch get_encoding_from_locale()");
#endif
	G_CONST_RETURN char *locale_encoding = NULL;
	if (setlocale(LC_CTYPE, locale)) g_get_charset(&locale_encoding);
	return g_strdup(locale_encoding);
}

gboolean check_string_in_array(gchar *str, gchar **lists)
{
#ifdef DETAIL
	g_debug("! Launch check_string_in_array() with str = %s", str);
#endif
	if ((str==NULL) || (lists==NULL))
		return FALSE;

	gint i=-1;

	while (lists[++i])
	{
		// g_debug("Checking %s and %s...", str, lists[i]);
		if (strcmp(str, lists[i])==0)
			return TRUE;
	}
	return FALSE;
}

gchar *get_proc_data(pid_t pid, gchar *file, gsize *length)
{
#ifdef FULL
	g_debug("! Launch get_proc_data() with pid = %d, file = %s", pid, file);
#endif
	if (! proc_exist) return NULL;
	if (pid<1) return NULL;
	
	gchar *contents=NULL;
	gint timeout=0;
	gchar *proc_path = g_strdup_printf("/proc/%d", (gint)pid);
	gchar *file_path = g_strdup_printf("%s/%s", proc_path, file);

	while (g_file_test(proc_path, G_FILE_TEST_EXISTS) &&
	       g_file_get_contents (file_path, &contents, length, NULL))
	{
		timeout++;
		// g_debug("The contents length is %d\n", *length);
		if (*length==0)
		{
			gsize len;
			gchar *stat = get_proc_data(pid, "stat", &len);
			if (len)
			{
				gchar **stats = split_string(stat, " ()", 6);
				if (stats && stats[4][0]=='Z')
				{
					g_warning("The child process \"(%s) %s\" has died. Abort.", stats[0], stats[2]);
					g_free(stat);
					g_strfreev(stats);

					g_free(contents);
					contents = NULL;
					break;
				}
				g_strfreev(stats);
			}
			g_free(stat);
			g_message("Waiting for /proc/%d/%s...", (gint)pid, file);
			// we should wait until "/proc/%d/file" is not empty
			// usleep(100000);
			sleep (1);
			// contents = "" here
			g_free(contents);
			contents = NULL;
		}
		else
			break;
		
		// we only try for 10 times
		if (timeout>=10)
		{
			g_warning("Failt when waiting for /proc/%d/%s. Abort.", (gint)pid, file);
			break;
		}
	}
	g_free(file_path);
	g_free(proc_path);
	// g_debug("contents = %s", contents);
	return contents;
}

// it will check if the count of returned string array is < max_tokens
gchar **split_string(gchar *str, gchar *split, gint max_tokens)
{
#ifdef FULL
	g_debug("! Launch split_string with str = %s, split = %s, max_tokens = %d", str, split, max_tokens);
#endif
	if (str==NULL) return NULL;

	// g_debug("contents = %s", contents);
	gchar **datas = g_strsplit_set(str, split, max_tokens);
	gint i = -1;

	// while (datas[++i])
	//	g_debug("Got data[%d] = %s", i, datas[i]);
	// g_debug("Got i = %d in split_string()", i);
	// i = -1;
	if (max_tokens>0)
	{
		while (datas[++i]);
		if (i<max_tokens)
		{
			g_strfreev(datas);
			datas = NULL;
		}
	}
	return datas;
}

gchar **get_pid_stat(pid_t pid, gint max_tokens)
{
#ifdef FULL
	g_debug("! Launch get_pid_stat with pid = %d, max_tokens = %d", pid, max_tokens);
#endif
	gsize length;
	gchar *stat = get_proc_data(pid, "stat", &length);
	gchar **stats = NULL;
	if (stat) stats = split_string(stat, " ()", max_tokens);
	g_free(stat);
	return stats;
}

// The returned string should be freed when no longer needed.
gchar *convert_text_to_html(gchar *text, gboolean free_text, gchar *color, gchar *tag, ...)
{
#ifdef DETAIL
	g_debug("! Launch convert_text_to_html() with text = %s, color = %s, tag = %s",
		text, color, tag);
#endif
	gchar *markup_escape_text = g_markup_escape_text(text, -1);

	if (color)
	{
		gchar *color_profile = g_strdup_printf ("<span foreground=\"%s\">%s</span>",
							color, markup_escape_text);
		g_free(markup_escape_text);
		markup_escape_text = color_profile;
	}

	va_list arg_ptr;
	va_start(arg_ptr, tag);
	while (tag)
	{
		gchar *new_html_string = g_strdup_printf("<%s>%s</%s>",
							 tag, markup_escape_text, tag);
		g_free(markup_escape_text);
		markup_escape_text = new_html_string;
		tag = va_arg(arg_ptr, char *);
	}
	va_end(arg_ptr);
	if (free_text) g_free(text);

	return markup_escape_text;
}

// The returned string should be freed when no longer needed.
gchar *join_strings_to_string(const gchar separator, const gint total, const gchar *string, ...)
{
#ifdef DETAIL
	g_debug("! Launch join_strings_to_string() with separator = %c, total = %d",
		separator, total);
#endif
	GString *strings = g_string_new ("");
	gint i;

	va_list arg_ptr;
	va_start(arg_ptr, string);
	for (i=0; i<total; i++)
	{
		if (string && string[0]!='\0')
		{
			if (strings->len)
				g_string_append_printf(strings, "%c%s", separator, string);
			else
				g_string_append_printf(strings, "%s", string);
		}
		// g_debug("[%2d] convert_strings_to_string(): strings = %s", i, strings->str);
		string = va_arg(arg_ptr, char *);
	}
	va_end(arg_ptr);
	// g_debug("convert_strings_to_string(): strings = %s", strings->str);
	
	return g_string_free(strings, FALSE);
}
